import bge
import time
from socket import *
import pickle,zlib
import os
import random 
from math import floor

cont    = bge.logic.getCurrentController()
own     = cont.owner

class clientState():
    
    def __init__(self, name="UClient",chunkSize = 8,seed="PuppyCat",requestChunks=[],state="alive"):
        self.seed           = seed
        self.state           = state
        self.name           = name
        self.chunkSize      = chunkSize
        self.requestChunks  = set(requestChunks)
        #print( seed )
    def compact(self):
        return {"state":self.state ,"name":self.name,"seed":self.seed,"chunkSize":self.chunkSize,"requestChunks":self.requestChunks}
    def expand(self,args):
        self.name,self.seed,self.chunkSize,self.requestChunks = args
    def updateRequestChunks(self,arr):
        currentArr  =  set(self.requestChunks)
        newArr      =  set(arr)
        resultArr   =  currentArr.union(newArr)
        self.requestChunks = resultArr


class client():
    def __init__(self, name="None",HOST='127.0.0.1',PORT=9999,BUFSIZ=8192,map_size=13,chunkSize=32,seed="None",viewDistance = 1):
        self.name   = name
        self.HOST   = HOST
        self.PORT   = PORT
        self.BUFSIZ = BUFSIZ
        self.map_size = map_size
        self.chunksToBeLoaded = set()
        self.chunksLoaded  = set()
        self.chunkSize = chunkSize
        self.seed = seed
        self.asyncToken = None
        self.viewDistance = viewDistance
        self.playerChunk = (0,0)
    def __del__(self):
        # in case of process close
        self.kill()                
    def kill(self):
        print(self.name,": Client killed.")
        try:
            # notifiy the server of the demise
            data = bytes("/kill","UTF-8")
            self.socket.send(data)
        except:
            print(self.name,": Server dead")
        time.sleep(1/10)
        self.socket.close()
        
    
    def loadChunks(self,playerLoc):

        if self.asyncToken:
            if  self.asyncToken.progress<1:
                return
            
        work = self.chunksToBeLoaded - self.chunksLoaded

        wD = self.viewDistance
        if wD ==0:
            wD = self.map_size
        spiralView = get_chunk_id(wD,wD,playerLoc)
        
        for chunk in spiralView:
            if chunk in work:
                break
        else: return
        if not len(work): return

        x,y = chunk

        path = bge.logic.expandPath('//')

        self.asyncToken = bge.logic.LibLoad(path+"map/result_{}_{}.blend".format(x,y), 'Scene', async=True ,verbose=False, load_scripts=False) 
        self.asyncToken.onFinish = self.onCompletion
        bge.logic.globalChunkDict[(int(x),int(y))] = getChunkDict("dict_{}_{}".format(x,y)) 

        self.chunksLoaded.add(chunk)
    def onCompletion(self,status):
        scene = bge.logic.getCurrentScene()
        name = status.libraryName.split("\\")[-1].split(".")[0]
        mType,x,y = name.split("_")
        #print(name,x,y)
        donorMesh   =   "wave_{}_{}".format(x,y)
        chunk = scene.addObject( donorMesh, own)
        size = bge.logic.chunkSize
        chunk.worldPosition      = (0,0,0)#(int(x)*size,int(y)*size,0)
        chunk.worldOrientation = (0,0,0)
        chunk.worldScale = (1,1,1)
        return True
    def run(self):
        
        ADDR        = (self.HOST, self.PORT)
        tcpCliSock  = socket(AF_INET, SOCK_STREAM)
        self.socket = tcpCliSock           
        
        while 1:
            try:
                tcpCliSock.connect(ADDR)   
            except:
                print("Connection Refused / No server.\n( Lags from retrying to connect )")                
                for i in range( 1 ): yield True  
                continue
            break
        
        self.socket.setblocking(0)
        
        mapX = mapY = self.map_size

        async_load_happening = False
       
        gameState = clientState(chunkSize = self.chunkSize,seed = self.seed)  
       
        COMPRESSION = True
        doneChunks = set()
        while 1:         
            currentAreaChunks =    set(get_chunk_id(mapX,mapY,self.playerChunk))
            playerChunks =  currentAreaChunks - doneChunks   
            self.chunksToBeLoaded = doneChunks & currentAreaChunks


            gameState.updateRequestChunks(playerChunks)
            data = pickle.dumps(gameState.compact())
            while 1:
                try: 
                    if COMPRESSION: data = zlib.compress(data) 
                    #print( data )         
                    tcpCliSock.send(data)                    
                except:
                    print("Can't send data, try again later")
                    for i in range(1): yield True
                    continue
                break
            #expect a response after ~~8 ms, 
            # might as well wait a logic tick
            yield True
            response = False          
            for try_to_recieve_ in range(330):
                try:            
                    response = tcpCliSock.recv(self.BUFSIZ)  
                    break                  
                except:
                    yield True                    
            if not response: continue 
            if COMPRESSION: response = zlib.decompress(response)
            response  =   pickle.loads(response)
            doneChunks = doneChunks.union(response["doneTasks"])
            bge.logic.chunkCount = len(doneChunks)
          
            yield True          


    
def get_chunk_id(X,Y,player_chunk):
    # spiraly expands the search net
    # can set the borders        
    x ,y = 0,0
    dx,dy = 0,-1  
    spiralSet = []
    for i in range(max(X,Y)**2):
        if (-X/2 < x <= X/2) and (-Y/2 < y <= Y/2):
            spiralSet.append( (x+player_chunk[0],y+player_chunk[1]) )
        if x == y or (x < 0 and x == -y) or (x > 0 and x == 1-y):
            dx, dy = -dy, dx
        x, y = x+dx, y+dy        
    return spiralSet

def getChunkDict(dictName):
    
    #local_path = os.path.abspath(__file__)
    #local_path =  local_path.replace("game_client.py","")
    #local_path =  local_path.replace("block_creation.blend","")

    local_path  = bge.logic.expandPath("//")
    local_path += "\\map\\"+dictName+".txt"
    
    d = pickle.load( open( local_path, "rb" ) )
         
    return d

def trunc(x,y,chunk_size):
    
    #funcion to create propper chunk id
    #works for negatice coordinates
    #should make use of math funcions though...    
    if x<0  :   x+=0.00000001
    if y<0  :   y+=0.00000001 
    c_x,c_y = int(x/chunk_size),int(y/chunk_size) 
    if x<0      :   c_x -= 1
    if y<0      :   c_y -= 1
    return (c_x,c_y)

def main():
    
    if not "client" in own:
        try: bge.logic.seed
        except: return
        own["loader"]           = client(
            name="Game",
            map_size=own["map_size"],
            chunkSize = own["chunkSize"],
            seed = bge.logic.seed,
            viewDistance = own["viewDistance"])
        own["client"]           = own["loader"].run()    
        bge.logic.chunkCount    = 0
       
        own["world_done"]       = False        
        own["time"]             = str(time.time())
        bge.logic.done          = False

        bge.logic.globalChunkDict   = {}
        bge.logic.chunkSize         = own["chunkSize"]
        bge.logic.manualVisibility  = 0
        bge.logic.trunc             = trunc
        scene                       = bge.logic.getCurrentScene()
        bge.logic.player            = scene.objects["Player"]        

        own["loader"].playerChunk   = (0,0)
        own["seed"] = bge.logic.seed
        
        print("\n","Starttt:",own["time"], "\n")


    own["ChunkCount"]   = bge.logic.chunkCount
    own["LoadCount" ]   = len(own["loader"].chunksLoaded)

    
    
    if own["updateLoading"]:
        playerLoc   = tuple(bge.logic.player.worldPosition.xy)
        playerChunk = trunc(playerLoc[0],playerLoc[1],bge.logic.chunkSize )       
        own["loader"].playerChunk = playerChunk
        
    own["loader"].loadChunks(own["loader"].playerChunk)   
        
    next(own["client"])



               
