import bpy
import math
from math import floor
import bmesh
import time 
import random
import os, sys
from mathutils import Vector
import pickle
from mapCreationLogic.perlin import Noise
from collections import defaultdict
# vector  = mathutils.Vector 

CHUNK_SIZE = 32
DELETE = 0
WORLD_SIZE = 1
MEM_OPT = 0
TEST_SEED = "CupCakes"
NO_COLLISION = True


def create_mesh(x,y, seed, chunkSize ):
    
    random.seed(seed)       
    bpy.ops.object.select_all(action="SELECT")        
    bpy.context.scene.cursor_location = (x*chunkSize, y*chunkSize, 0.0)
    
    #print( bpy.context.scene.layers )
    bpy.context.scene.layers[3] = True
    bpy.context.scene.layers[0] = False
    
    # set the file path and file names
    file_path           = (bpy.path.abspath("//"))
    file_directory      = os.path.dirname(file_path)
    output_path         = os.path.join(file_directory, "map/result_{}_{}.blend".format(x,y))    
    mesh_name           = "wave_{}_{}".format(x,y)
    
    # create the chunk
    obj = createChunk( Argx = int(x), Argy = int(y) , size =  chunkSize, name = mesh_name , seed = seed)
    if not obj: return

    ret = bpy.ops.wm.save_as_mainfile( check_existing   = False,
                                       filepath         = output_path,
                                       compress         = True,
                                       use_mesh_compat  = True ,
                                       copy             = True) 

    obj.name = "initial"



def exampleCreator(objName):
    """ Creates a uv & vert list of a reference object """
    exampleCube         = bpy.data.objects[objName]
    exampleCubeMesh     = exampleCube.data 
    exampleCubeFaces    = []
    exampleCubeUV       = []
    
    tempM               = bmesh.new()
    tempM.from_mesh(      exampleCubeMesh  )
    uvLayer             = tempM.loops.layers.uv.verify()   

    for face in tempM.faces:
        uvs ,verts  = [], []            
        for f in face.loops:                
            uvs.append( f[uvLayer].uv.copy())
        for v in face.verts:
            verts.append(v.co.copy())
        exampleCubeFaces.append(   verts )
        exampleCubeUV.append( uvs )      
            
    tempM.free()
    return (exampleCubeFaces,exampleCubeUV)    

def createChunk( Argx = 0, Argy = 0 , size = 32 , name = "None", seed = "Butter cream."):

    #print( "meshCreator seed:",seed )

    """ Initialize the mesh to be modified """

    scene = bpy.context.scene   
    if not "initial" in scene.objects: return None    
    obj =  scene.objects["initial"]    
    obj.select = True   

    # create bmesh
    bm = bmesh.new()      
    
    # initialize mesh   
    uvLayer = bm.loops.layers.uv.verify()   
    
    bm.faces.layers.tex.verify()
    #mat_name = 'TexMat'

    #use the material on the seed object.
    #which can be a duplicate of the example objects in builds its self from
    #if not len(bpy.data.materials): create_mat(mat_name)   
    #obj.data.materials.append(bpy.data.materials[mat_name])
    
    material_id = len(obj.data.materials) - 1         
    

    """ Define the donor/example/reference cubes to be faked """
    exampleCubes = [    exampleCreator("grass_dirt"),
                        exampleCreator("dirt"),
                        exampleCreator("stone")]

    exampleCubes = [    exampleCreator("Stones"),
                        exampleCreator("Concrete"),
                        exampleCreator("Wood"),
                        exampleCreator("Pellet"),
                        exampleCreator("Tiles")]

    """ Initialize the data containers for the mesh """   
    # block : face : [verts]
    block_dict = {} 

    # block(x,y) : set(range(-256,this.z)) # check this    
    elevation_dict = defaultdict(lambda:set())

    # helper container for keys
    dictKeys = set() 

    #off sets
    loc_x,loc_y = Argx*size,Argy*size
     
    # initialize noise generator    
    noise  = Noise(8, seed)      

    #################################################

    """ Create the chunks structure/features """
    from mapCreationLogic  import  DisIntegrateBuilding


    height      = 5
    rooms       = 1
    room_size   = int(size/rooms)    
    
    roomDict = {}
    for dx in range(0,rooms):
        for dy in range(0,rooms):
            room = DisIntegrateBuilding.generate(0,0,room_size,height)
            for x,y,z,block in room:
                loc = (loc_x+x+dx*room_size,loc_y+y+dy*room_size)
                roomDict[(loc[0],loc[1],z)] = block
                dictKeys.add((loc))                
                elevation_dict[loc].add(z)
    
    dictKeys = list(dictKeys )
    
    ####################################################

    faces   = bm.faces
    verts   = bm.verts
    newVert = verts.new
    newFace = faces.new
   
    #delta  =  [vector((-1,0,0)),vector((0,1,0)),vector((1,0,0)),vector((0,-1,0)),vector((0,0,-1)),vector((0,0,1))]   
    delta   =  [(-1,0,0),(0,1,0),(1,0,0),(0,-1,0),(0,0,-1),(0,0,1)]   
    index   = 0
    
    def ground_type( height,top):
        #help function
        if height < 1:
            return 2
        if height == top:
            return 0
        return 1

    for x,y in dictKeys:        

        maxH            = max(elevation_dict[(x,y)])
        elevationDict   = elevation_dict[(x,y)]

        for h in elevationDict:                      
            
            #tempDict = { (-1,0,0):[],(0,1,0):[], (1,0,0):[],(0,-1,0):[],(0,0,-1):[],(0,0,1):[],0:ground_type(h,maxH)}
            blockType           = ground_type(h,maxH) 
            blockType           = roomDict[(x,y,h)]
            tempDict            = { 0:blockType}
            posVect             = Vector(  (x,y,h)  ) 
            # get the example features
            exampleUv       = exampleCubes[blockType][1]
            exampleFaces    = exampleCubes[blockType][0]
            
            
            tempList = range(6)   
            """ Disolve the parts obstructed from view """        
            #"""
            tempList = [] 
            if not h-1 in elevationDict :                    
                tempList.append(4)
            if not  h+1 in elevationDict:  
                tempList.append(5)  
            if  not  h in elevation_dict[(x-1,y)]:
                tempList.append(0)
            if  not  h in elevation_dict[(x,y+1)]:
                tempList.append(1)
            if  not  h in elevation_dict[(x+1,y)]:
                tempList.append(2)
            if  not  h in elevation_dict[(x,y-1)]:
                tempList.append(3)
            #"""
    
            for tempIndex in tempList:
                
                # get the faces and uvs
                faceDeltaList   = exampleFaces[tempIndex]  
                faceUvList      = exampleUv[tempIndex]
                vertAmount      = len(faceDeltaList) # quads or tris face
                
                # save the vertex index in the block:face:verts
                tempDict[ delta[tempIndex] ] = range(index,index+vertAmount)
                index += vertAmount
                
                face = newFace( [ newVert( tuple(posVect+faceDelta)  ) for faceDelta in  faceDeltaList  ])               
                
                for i, loop in enumerate(face.loops):
                    loop[uvLayer].uv = faceUvList[i] 

            block_dict[ (x,y,h) ] = tempDict

    """ generate free faces """    
    extraFaces = int(len( block_dict )*0.05)
    #print("blockDlen", len( block_dict),extraFaces)
    freeFaces = []
    for i in range(extraFaces):
        # get the faces and uvs
        faceDeltaList   = exampleFaces[0]  
        faceUvList      = exampleUv[0]
        vertAmount      = len(faceDeltaList) # quads or tris face  

        freeFaces.append( range(index,index+vertAmount) )      

        index += vertAmount
        
        posVect         = Vector(  (loc_x,loc_y,-10000)  ) 
        face = newFace( [ newVert( tuple(posVect+faceDelta) ) for faceDelta in faceDeltaList])               
        
        for i, loop in enumerate(face.loops):
            loop[uvLayer].uv = faceUvList[i] 

    block_dict["freeFaces"] = freeFaces


####################################################             
    
    # save the vertex data to a text file dump
    create_dict_dump(block_dict,Argx,Argy)   
    
    # prepare the mesh for shipping
    me = obj.data
    bm.to_mesh(me)    
    bm.free()    
    
    if NO_COLLISION: obj.game.physics_type = "NO_COLLISION"
    
    obj.name        = name
    obj.data.name   = name
    #print("obj name", obj.name, obj.data.name)

    return obj
    
    
            
def create_dict_dump(block_dict,x,y):
    """
    Creates dict dump in a file
    """ 
    file_path = (bpy.path.abspath("//"))
    file_directory = os.path.dirname(file_path)
    output_path = os.path.join(file_directory, "map\dict_{}_{}.txt".format(x,y))    
    pickle.dump( block_dict , open( output_path, "wb" ) )



def create_mat(mat_name):
    
    # Load image file. Change here if the snippet folder is 
    # not located in you home directory.
    
    picture = "\\tile_texture_2.png"
    local_path = os.path.abspath(__file__)
    local_path =  local_path.replace("\\"+local_path.split("\\")[-1],"")
    local_path =  local_path.replace("\\"+local_path.split("\\")[-1],"")
    local_path += picture
    
    
    realpath = local_path
    try:
        img = bpy.data.images.load(realpath)
    except:
        raise NameError("Cannot load image %s" % realpath)
 
    # Create image texture from image
    cTex = bpy.data.textures.new('ColorTex', type = 'IMAGE')
    cTex.image = img
    
    # Create material
    mat = bpy.data.materials.new(mat_name)
    mat.specular_intensity = 0.0
    # Add texture slot for color texture
    mtex = mat.texture_slots.add()
    mtex.texture = cTex
    mtex.texture_coords = 'UV'
    mtex.use_map_color_diffuse = True 
    mtex.use_map_color_emission = True 
    mtex.emission_color_factor = 0.5
    mtex.use_map_density = True 
    mtex.mapping = 'FLAT' 




def flatten(n):
    #Turns the smooth noise 
    #into discrete height layers
    if n<0:
        n*= 1+abs(n)*5
    s = 25
    z = int(n*s)   
    return z



