import bge
from time import time,sleep
from socket import *
import os,sys
import random 
from mathutils import Vector
from math import sin


sys.path.append ( os.path.abspath(__file__).replace( os.path.basename(__file__),"")) 
sys.path.append ( bge.logic.expandPath("//") ) 

from structures import *
from serverSettings import *
import openServerFromGame

cont    = bge.logic.getCurrentController()
own     = cont.owner



class client():
    def __init__(self, name="None",HOST='127.0.0.1',PORT=9999,BUFSIZ=1024):
        self.name   = name
        self.HOST   = HOST
        self.PORT   = PORT
        self.BUFSIZ = BUFSIZ
        self.shotOffset = 0    


    def __del__(self):
        # in case of process close
        self.kill()   

    def kill(self):
        print(self.name,": Client killed.")
        self.socket.close() 
 
    def myGameData(self,health = 1):
        player  = bge.logic.player
        lin     = vector3( *(player.worldLinearVelocity).to_tuple())
        ang     = vector3( *(player.worldAngularVelocity).to_tuple())
        tra     = vector16( *([ i  for v in player.worldTransform for i in v.to_tuple() ]))

        players = [ PlayerStruct(  self.myID, lin, ang, tra) ]

        shots   = []        
        for index,(v,loc) in enumerate(bge.logic.shots):
            shots.append( Shot(self.myID+(self.shotOffset+index)%self.mySpace,vector3(*v),vector3(*loc)))
        self.shotOffset = (self.shotOffset + len(shots))%self.mySpace
        bge.logic.shots = []
        
        return packPacket(players,shots)   

    
    def run(self):
        
        ADDR        = (self.HOST, self.PORT)
        tcpCliSock  = socket(AF_INET, SOCK_STREAM)
        self.socket = tcpCliSock           
        myGameData  = self.myGameData

        while 1:
            failed = False
            try:
                tcpCliSock.connect(ADDR)
                break   
            except:
                failed = True             
                              
            yield True               
            
       
        self.socket.setblocking(0)
        sleep(0.15)
        yield True   
        

        response    = tcpCliSock.recv(self.BUFSIZ)        
        welcome     = Welcome.from_buffer_copy(response)
        self.myID   = welcome.id
        self.myDomain = list(range(self.myID,self.myID+welcome.space))
        self.mySpace = welcome.space
        bge.logic.seed = welcome.seed
        print( "\nnreceived seed:" ,welcome.seed,"\n")

        tcpCliSock.send(myGameData())                    
        yield True


        while 1:         
                      
            try:            
                response = tcpCliSock.recv(self.BUFSIZ)                                    
            except:
                yield
                continue

            own["dataLoad"] = sys.getsizeof(response) 
            players,shots   = translate(response)
            own["ids"]      = str([i.id for i in players])
            own["shots"]    = str([len(shots),[shot.id for shot in shots]])
            own["players"]  = len(players)

            updatePlayers(players,self.myID)
            updateShots(shots,self.myDomain)
            

            gameState = myGameData()
            while 1:
                try:                             
                    tcpCliSock.send(gameState)
                    break                    
                except:
                    print("Can't send data, try again later")
                    yield True                                    

            yield True          

def beamCalculation(pos,aimPos,aimVect):

    A = Vector(list(aimPos)) 
    D = Vector(list(aimVect)) 
    P = pos
    B = P-A
    X = A + (P-A).dot(D)*D
    C = P-X
    N = C.normalized()
    L = C.length

    angle = D.angle(B,666)     

    forcePower = min(50,max(0, 1/(L+1)**2 * 100 ) )
    if  D.angle(B,666)  > 1.5707963705062866: forcePower = 0

    force = (N+D/2)* forcePower

    return (force,forcePower)

def updatePlayers(players,myID):
    me          = bge.logic.player
    mePos       = me.worldPosition
    dictionary  = bge.logic.players
    what        = "newPlayer"
    
    for player in players:
        pName         = int(player.id)

        if pName == myID: continue        

        if not pName in dictionary and pName>=0:
            dictionary[pName] = bge.logic.scene.addObject("newPlayer",own)

        elif pName < 0:            
            if abs(pName) in dictionary:
                dictionary[abs(pName)].endObject()            
                del dictionary[abs(pName)]
            continue

        playerObj = dictionary[pName]
        playerObj.worldLinearVelocity  = (player.worldLinearVelocity)
        playerObj.worldAngularVelocity = (player.worldAngularVelocity)        

        worldTransform = [ player.worldTransform[i:i+4] for i in range(0,16,4)]
        playerObj.worldTransform       = worldTransform

def updateShots(shots,myDomain):
    me          = bge.logic.player
    mePos       = me.worldPosition
    dictionary  = bge.logic.serverShots
    what        = "Beam"
    lifeTime    = 2
    for shot in shots:
       
        key         = int(shot.id)

        if not key in dictionary and key>=0:
            dictionary[key] = [shot,None,time()]   
            visual = bge.logic.scene.addObject(what,own,20)   

            visual.worldPosition = shot.pos
            visual.alignAxisToVect((shot.vector), 1)

            if key in myDomain: continue

            forceVect,power = beamCalculation(mePos,shot.pos,shot.vector)        
            if power < 2: continue
            me["knockOut"] = 60                    
            me.setLinearVelocity(forceVect)


def registerShot():
    mouse       = bge.logic.mouse.events
    leftClick   = bge.events.LEFTMOUSE
    JUST_ACTIVATED = bge.logic.KX_INPUT_JUST_ACTIVATED
    
    if mouse[leftClick]:
        if not "shotTimer" in own:
            own["shotTimer"] = str(time()-1)           
        last = float(own["shotTimer"])
        if time()-last>  1  : #timelimit before reshot allowed
            
            own["shotTimer"] = str(time())
            #v = own.parent.getVectTo( own )[1].normalized()
            v = - bge.logic.playerCamera.getScreenVect(0.5, 0.5)
            try:
                if len(bge.logic.shots) < 8:
                    bge.logic.shots.append( (v,bge.logic.playerCamera.worldPosition) )
            except:
                pass  



def main():

    CONNECT_TO_IP = SERVER_IP

    if not "connection" in own:         
        scene                       = bge.logic.getCurrentScene()
        bge.logic.scene             = scene
        bge.logic.player            = scene.objects["Player"]
        bge.logic.playerCamera      = bge.logic.player.children["Camera"]
        bge.logic.players           = { }
        bge.logic.shots             = []
        bge.logic.serverShots       = {}

        
        givenServerUp = openServerFromGame.isServerOpen(SERVER_IP,PORT)
        print("test if game server up",CONNECT_TO_IP, givenServerUp )
        if not givenServerUp:
            myComputerIp =  gethostbyname(gethostname())
            myComputerServerUp = openServerFromGame.isServerOpen(myComputerIp,PORT)
            print("test if game server up on this computer",myComputerIp, myComputerServerUp )
            if not myComputerServerUp:       
                print( "open server")
                openServerFromGame.launch()
            CONNECT_TO_IP = myComputerIp
            
        own["HOST"]                 = CONNECT_TO_IP
        own["connection"]           = client(HOST = CONNECT_TO_IP,PORT = PORT,BUFSIZ = BUFSIZ).run()  
           


                       
        

    
    next(own["connection"])

    
    registerShot()


 



               
