from ctypes import *
import random
import zlib
dataType = c_byte
dataType = c_short
dataType = c_float

COMPRESSION = 1
vector3     = dataType * 3
vector16    = dataType * 16

class PlayerStruct(Structure):
    _fields_ = [("id", c_short),                
                ("worldLinearVelocity", dataType * 3),
                ("worldAngularVelocity", dataType * 3),
                ("worldTransform", dataType * 16)
                ]

class Header(Structure):
    _fields_ = [        
        ("players", c_ubyte),                
        ("shots",   c_ubyte)           
    ]

class Shot(Structure):
    _fields_ = [       
        ("id",      c_short), 
        ("vector",  c_float * 3 ),                
        ("pos",     c_float * 3 )            
    ]

class Welcome(Structure):
    _fields_ = [       
        ("id",      c_short),                 
        ("space",     c_ubyte ),
        ("seed",     c_wchar*16 )


    ]


def ran(limit,size):
    r = random.randint
    return ([ r(0,limit) for i in range(size)])

def translate(data):
    if COMPRESSION: data = zlib.decompress(   data  )

    players = []
    shots   = []

    header      = Header.from_buffer_copy(data)
    headerSize  = sizeof(Header)
    pStructSize = sizeof(PlayerStruct)
    pDataSize   = pStructSize * header.players

    for pIndex in range(headerSize,pDataSize+headerSize,pStructSize):
        player = PlayerStruct.from_buffer_copy(data,pIndex)
        players.append( player )

    shotSize    = sizeof(Shot)
    offset      = headerSize + pDataSize
    sDataSize   = shotSize * header.shots

    for sIndex in range(offset,offset+sDataSize,shotSize):
        shots.append( Shot.from_buffer_copy(data,sIndex))

    #print( "shots:",len(shots),header.shots)
    return [players,shots]

def packPacket(players,shots):
    playerCount = len(players)
    shotCount   = len(shots)

    playerArray = PlayerStruct * playerCount
    players     = playerArray( *players )

    shotArray   = Shot * shotCount
    shots       = shotArray( *shots)

    header      = Header( playerCount, shotCount )
    packet      = bytearray(header) + bytearray(players) +bytearray(shots)

    if COMPRESSION: packet = zlib.compress(packet)
    return packet


def vectors():
    return [ vector3( *ran(100,3)),vector3( *ran(100,3)),vector16( *ran(100,16)) ]     
